/*
 * Copyright 1990 - 1994, Julianne Frances Haugh
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * chpasswd - update passwords in batch
 *
 *      chpasswd reads standard input for a list of colon separated
 *      user names and new passwords.  the appropriate password
 *      files are updated to reflect the changes.  because the
 *      changes are made in a batch fashion, the user must run
 *      the mkpasswd command after this command terminates since
 *      no password updates occur until the very end.
 *
 * 1997/07/29: Modified to take "-e" argument which specifies that
 *             the passwords have already been encrypted.
 *             -- Jay Soffian <jay@lw.net>
 */

#include <config.h>

#include "rcsid.h"
RCSID (PKG_VER "$Id: chpasswd.c,v 1.29 2005/08/11 16:23:34 kloczek Exp $")
#include <fcntl.h>
#include <getopt.h>
#include <pwd.h>
#include "pam_chpw.h"
#include <stdio.h>
#include <stdlib.h>
#ifdef USE_PAM
#include <security/pam_appl.h>
#include <security/pam_misc.h>
#endif				/* USE_PAM */
#ifdef SHADOWTCB
#include "tcbfuncs.h"
#endif
#include "prototypes.h"
#include "defines.h"
#include "pwio.h"
#include "nscd.h"
#include "shadowio.h"
/*
 * Global variables
 */
static char *Prog;
static int eflg = 0;
static int md5flg = 0;

/* local function prototypes */
static void usage (void);

/*
 * usage - display usage message and exit
 */

static void usage (void)
{
	fprintf (stderr, _("Usage: chpasswd [options]\n"
			   "\n"
			   "Options:\n"
			   "  -e, --encrypted	supplied passwords are encrypted\n"
			   "  -h, --help		display this help message and exit\n"
			   "  -m, --md5		use MD5 encryption instead DES when the supplied\n"
			   "			passwords are not encrypted\n"));
	exit (1);
}

static int paste_pwd_shadow (char *name, char *pwd)
{
	const struct spwd *sp;
	struct spwd newsp;
	long now = time ((long *) 0) / (24L * 3600L);

	if (!tcb_user (name))
		return 0;
	if (!spw_lock ())
	{
		fprintf (stderr, "can't lock shadow file for %s\n",
			name);
		return 0;
	}
	if (!spw_open (O_RDWR))
	{
		fprintf (stderr, "can't open shadow file for %s\n",
			name);
		spw_unlock ();
		return 0;
	}
	sp = spw_locate (name);
	if (sp)
	{
		newsp = *sp;
		newsp.sp_pwdp = pwd;
		newsp.sp_lstchg = now;
	}
	else
	{
		fprintf (stderr, "can't locate shadow entry for %s\n", name);
		return 0;
	}
	if (!spw_update (&newsp))
	{
		fprintf (stderr, "can't update shadow entry for %s\n", name);
		return 0;
	}
	if (!spw_close ())
	{
		fprintf (stderr, "error updating shadow file\n");
		return 0;
	}
	spw_unlock ();
	return 1;
}

static int paste_pwd (char *name, char *pwd)
{
	const struct passwd *pw;
	struct passwd newpw;
#ifdef ATT_AGE
	long now = time ((long *) 0) / (24L * 3600L);
#endif
	if (spw_file_present ())
		return paste_pwd_shadow (name, pwd);

	if (!pw_lock ())
	{
		fprintf (stderr, "can't lock password file\n");
		return 0;
	}
	if (!pw_open (O_RDWR))
	{
		fprintf (stderr, "can't open password file\n");
		return 0;
	}

	pw = pw_locate (name);
	if (!pw)
	{
		fprintf (stderr, "unknown user %s\n",
			name);
		return 0;
	}
	newpw = *pw;
	newpw.pw_passwd = pwd;
#ifdef ATT_AGE
	if (newpw.pw_age[0])
	{
		strcpy (newage, newpw.pw_age);
		strcpy (newage + 2, l64a (now / 7));
		newpw.pw_age = newage;
	}
#endif

	if (!pw_update (&newpw))
	{
		fprintf (stderr, "cannot update password entry\n");
		return 0;
	}
	if (!pw_close ())
	{
		fprintf (stderr, "error updating password file\n");
		return 0;
	}
	pw_unlock ();
	return 1;
}

#ifdef USE_PAM
static struct pam_conv conv = {
	misc_conv,
	NULL
};
#endif				/* USE_PAM */

int main (int argc, char **argv)
{
	char buf[BUFSIZ];
	char *name;
	char *cp;

	static int errors = 0;
	int line = 0;

#ifdef USE_PAM
	pam_handle_t *pamh = NULL;
	struct passwd *pampw;
	int retval;
#endif

	Prog = Basename (argv[0]);

	setlocale (LC_ALL, "");
	bindtextdomain (PACKAGE, LOCALEDIR);
	textdomain (PACKAGE);

	{
		int option_index = 0;
		int c;
		static struct option long_options[] = {
			{"encrypted", no_argument, NULL, 'e'},
			{"help", no_argument, NULL, 'h'},
			{"md5", no_argument, NULL, 'm'},
			{NULL, 0, NULL, '\0'}
		};

		while ((c =
			getopt_long (argc, argv, "ehm", long_options,
				     &option_index)) != -1) {
			switch (c) {
			case 'e':
				eflg = 1;
				break;
			case 'h':
				usage ();
				break;
			case 'm':
				md5flg = 1;
				break;
			case 0:
				/* long option */
				break;
			default:
				usage ();
				break;
			}
		}
	}

#ifdef USE_PAM
	retval = PAM_SUCCESS;

	pampw = getpwuid (getuid ());
	if (pampw == NULL) {
		retval = PAM_USER_UNKNOWN;
	}

	if (retval == PAM_SUCCESS) {
		retval = pam_start ("chpasswd", pampw->pw_name, &conv, &pamh);
	}

	if (retval == PAM_SUCCESS) {
		retval = pam_authenticate (pamh, 0);
		if (retval != PAM_SUCCESS) {
			pam_end (pamh, retval);
		}
	}

	if (retval == PAM_SUCCESS) {
		retval = pam_acct_mgmt (pamh, 0);
		if (retval != PAM_SUCCESS) {
			pam_end (pamh, retval);
		}
	}

	if (retval != PAM_SUCCESS) {
		fprintf (stderr, _("%s: PAM authentication failed\n"), Prog);
		exit (1);
	}
#endif				/* USE_PAM */

	/*
	 * Read each line, separating the user name from the password. The
	 * password entry for each user will be looked up in the appropriate
	 * file (shadow or passwd) and the password changed. For shadow
	 * files the last change date is set directly, for passwd files the
	 * last change date is set in the age only if aging information is
	 * present.
	 */

	while (fgets (buf, sizeof buf, stdin) != (char *) 0) {
		line++;
		if ((cp = strrchr (buf, '\n'))) {
			*cp = '\0';
		} else {
			fprintf (stderr, _("%s: line %d: line too long\n"),
				 Prog, line);
			errors++;
			break;
		}

		/*
		 * The username is the first field. It is separated from the
		 * password with a ":" character which is replaced with a
		 * NUL to give the new password. The new password will then
		 * be encrypted in the normal fashion with a new salt
		 * generated, unless the '-e' is given, in which case it is
		 * assumed to already be encrypted.
		 */

		name = buf;
		if ((cp = strchr (name, ':'))) {
			*cp++ = '\0';
		} else {
			fprintf (stderr,
				 _("%s: line %d: missing new password\n"),
				 Prog, line);
			errors++;
			break;
		}
		if (eflg) {
			if (!paste_pwd (name, cp)) {
				fprintf (stderr, "%s: line %d: unable to paste new hash\n",
					Prog, line);
				errors++;
				break;
			}
		} else if (!do_pam_chpass ("chpasswd", name, cp)) {
			fprintf (stderr, "%s: line %d: unable to change password for %s\n",
				Prog, line, name);
			errors++;
			break;
		}
	}
	
	if (errors)
		return 1;

	return (0);
}
