/*
  vipw, vigr  edit the password or group file
  with -s will edit shadow or gshadow file
 
  Copyright (C) 1997 Guy Maor <maor@ece.utexas.edu>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  */

#include <config.h>

#include "rcsid.h"
RCSID (PKG_VER "$Id: vipw.c,v 1.14 2005/08/03 16:00:46 kloczek Exp $")
#include "defines.h"
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>
#include <utime.h>
#ifdef SHADOWTCB
#include <tcb.h>
#include "tcbfuncs.h"
#endif
#include "getdef.h"
#include "commonio.h"
#include "prototypes.h"
#include "pwio.h"
#include "shadowio.h"
#include "groupio.h"
#include "sgroupio.h"
#include "nscd.h"
static const char *progname, *filename, *fileeditname;
static int filelocked = 0, createedit = 0;
static int securemode = 0;
static char *user = NULL;
static int (*unlock) (void);

/* local function prototypes */
static int create_backup_file (FILE *, const char *, struct stat *);
static void vipwexit (const char *, int, int);
static void vipwedit (const char *, int (*)(void), int (*)(void));

static int create_backup_file (FILE * fp, const char *backup, struct stat *sb)
{
	struct utimbuf ub;
	FILE *bkfp;
	int c;
	mode_t mask;

	mask = umask (077);
	bkfp = fopen (backup, "w");
	umask (mask);
	if (!bkfp)
		return -1;

	c = 0;
	if (fseeko (fp, 0, SEEK_SET) == 0)
		while ((c = getc (fp)) != EOF) {
			if (putc (c, bkfp) == EOF)
				break;
		}
	if (c != EOF || ferror (fp) || fflush (bkfp)) {
		fclose (bkfp);
		unlink (backup);
		return -1;
	}
	if (fclose (bkfp)) {
		unlink (backup);
		return -1;
	}

	ub.actime = sb->st_atime;
	ub.modtime = sb->st_mtime;
	if (utime (backup, &ub) ||
	    chmod (backup, sb->st_mode) ||
	    chown (backup, sb->st_uid, sb->st_gid)) {
		unlink (backup);
		return -1;
	}
	return 0;
}

#ifdef SHADOWTCB
static int prep_new(char **to_rename, char *fileedit, const char *file)
{
	FILE *f;
	struct stat st;

	if (!(f = fopen(fileedit, "r"))) return 0;
	if (unlink(fileedit)) return 0;
	if (!s_drop_priv()) return 0;
	if (stat(file, &st)) return 0;
	asprintf(to_rename, "%s+", file);
	if (!*to_rename) {
		fclose(f);
		return 0;
	}
	if (create_backup_file(f, *to_rename, &st)) return 0;

	return 1;
}
#endif

static void vipwexit (const char *msg, int syserr, int ret)
{
	int err = errno;

	if (createedit)
		unlink (fileeditname);
	if (filelocked)
		(*unlock) ();
	if (msg)
		fprintf (stderr, "%s: %s", progname, msg);
	if (syserr)
		fprintf (stderr, ": %s", strerror (err));
	fprintf (stderr, _("\n%s: %s is unchanged\n"), progname, filename);
	exit (ret);
}

#ifndef DEFAULT_EDITOR
#define DEFAULT_EDITOR "vi"
#endif

#define SCRATCHDIR ":tmp"

static void
vipwedit (const char *file, int (*file_lock) (void), int (*file_unlock) (void))
{
	const char *editor;
	pid_t pid;
	struct stat st1, st2;
	int status;
	FILE *f;
	char filebackup[1024], fileedit[1024];
	char *to_rename;

	snprintf (filebackup, sizeof filebackup, "%s-", file);
#ifdef SHADOWTCB
	if (securemode) {
		if (mkdir(TCB_DIR "/" SCRATCHDIR, 0700) && errno != EEXIST) {
			fprintf(stderr, "%s when trying to mkdir " TCB_DIR "/" SCRATCHDIR
				"\nare you sure you're the admin here? :^)\n", strerror(errno));
			exit(1);
		}
		snprintf(fileedit, sizeof fileedit, TCB_DIR "/" SCRATCHDIR "/.vipw.shadow.%s", user);
	} else
#endif
		snprintf(fileedit, sizeof fileedit, "%s.edit", file);
	unlock = file_unlock;
	filename = file;
	fileeditname = fileedit;

#ifdef SHADOWTCB
	if (securemode && !s_drop_priv()) {
		fprintf(stderr, "Unable to open %s\n", file);
		exit(1);
	}
#endif
	if (access (file, F_OK))
		vipwexit (file, 1, 1);
#ifdef SHADOWTCB
	if (securemode && !s_gain_priv()) {
		fprintf(stderr, "Unable to gain privs\n");
		exit(1);
	}
#endif
	if (!file_lock ())
		vipwexit (_("Couldn't lock file"), errno, 5);
	filelocked = 1;
#ifdef SHADOWTCB
	if (securemode && !s_drop_priv()) {
		fprintf(stderr, "Unable to open %s\n", file);
		exit(1);
	}
#endif
	/* edited copy has same owners, perm */
	if (stat (file, &st1))
		vipwexit (file, 1, 1);
	if (!(f = fopen (file, "r")))
		vipwexit (file, 1, 1);
#ifdef SHADOWTCB
	if (securemode && !s_gain_priv()) {
		fprintf(stderr, "Unable to gain privs\n");
		exit(1);
	}
#endif
	if (create_backup_file (f, fileedit, &st1))
		vipwexit (_("Couldn't make backup"), errno, 1);
	createedit = 1;

	editor = getenv ("VISUAL");
	if (!editor)
		editor = getenv ("EDITOR");
	if (!editor)
		editor = DEFAULT_EDITOR;

	if ((pid = fork ()) == -1)
		vipwexit ("fork", 1, 1);
	else if (!pid) {
		/* use the system() call to invoke the editor so that it accepts
		   command line args in the EDITOR and VISUAL environment vars */
		char *buf;

		buf = (char *) malloc (strlen (editor) + strlen (fileedit) + 2);
		snprintf (buf, strlen (editor) + strlen (fileedit) + 2,
			  "%s %s", editor, fileedit);
		if (system (buf) != 0) {
			fprintf (stderr, "%s: %s: %s\n", progname, editor,
				 strerror (errno));
			exit (1);
		} else
			exit (0);
	}

	for (;;) {
		pid = waitpid (pid, &status, WUNTRACED);
		if (WIFSTOPPED (status)) {
			kill (getpid (), SIGSTOP);
			kill (getpid (), SIGCONT);
		} else
			break;
	}

	if (pid == -1 || !WIFEXITED (status) || WEXITSTATUS (status))
		vipwexit (editor, 1, 1);

	if (stat (fileedit, &st2))
		vipwexit (fileedit, 1, 1);
	if (st1.st_mtime == st2.st_mtime)
		vipwexit (0, 0, 0);

	/*
	 * XXX - here we should check fileedit for errors; if there are any,
	 * ask the user what to do (edit again, save changes anyway, or quit
	 * without saving). Use pwck or grpck to do the check.  --marekm
	 */

	createedit = 0;
#ifdef SHADOWTCB
	if (securemode) {
		if (!prep_new(&to_rename, fileedit, file)) {
			fprintf(stderr, _("%s: can't restore %s: %s (your changes are in %s)\n"),
				progname, file, strerror(errno), fileedit);
			vipwexit(0,0,1);
		}
	} else
#endif
		to_rename = fileedit;
	unlink (filebackup);
	link (file, filebackup);
	if (rename (to_rename, file) == -1) {
		fprintf (stderr,
			 _
			 ("%s: can't restore %s: %s (your changes are in %s)\n"),
			 progname, file, strerror (errno), fileedit);
		vipwexit (0, 0, 1);
	}
#ifdef SHADOWTCB
	if (securemode && !s_gain_priv()) {
		fprintf(stderr, "Unable to gain privs\n");
		exit(1);
	}
#endif
	(*file_unlock) ();
}

static void usage(void)
{
#ifdef SHADOWTCB
	if (getdef_bool("USE_TCB"))
		fprintf(stderr,
"Usage:\n"
"`vipw' edits /etc/passwd        `vipw -s user' edits /etc/tcb/user/shadow\n"
"`vigr' edits /etc/group         `vigr -s' edits /etc/gshadow\n");
	else
#endif
		fprintf(stderr,
"Usage:\n"
"`vipw' edits /etc/passwd        `vipw -s' edits /etc/shadow\n"
"`vigr' edits /etc/group         `vigr -s' edits /etc/gshadow\n");
}

extern struct commonio_db shadow_db;

int main (int argc, char **argv)
{
	int flag;
	int editshadow = 0;
	char *c;
	int e = 1;
	int do_vipw;

	setlocale (LC_ALL, "");
	bindtextdomain (PACKAGE, LOCALEDIR);
	textdomain (PACKAGE);

	progname = ((c = strrchr (*argv, '/')) ? c + 1 : *argv);
	do_vipw = (strcmp (progname, "vigr") != 0);

	while ((flag = getopt (argc, argv, "ghps")) != EOF) {
		switch (flag) {
		case 'g':
			do_vipw = 0;
			break;
		case 'h':
			e = 0;
		case 'p':
			do_vipw = 1;
			break;
		case 's':
			editshadow = 1;
			break;
		default:
			usage();
			exit (e);
		}
	}
#ifdef SHADOWTCB
	if (do_vipw && editshadow && getdef_bool("USE_TCB")) {
		securemode = 1;
		user = argv[optind];
		if (!user) {
			usage();
			exit(1);
		}
		if (!tcb_user(user))
			exit(1);
	}
#endif
	if (do_vipw) {
		if (editshadow)
			vipwedit (shadow_db.filename, spw_lock, spw_unlock);
		else
			vipwedit (PASSWD_FILE, pw_lock, pw_unlock);
	} else {
#ifdef SHADOWGRP
		if (editshadow)
			vipwedit (SGROUP_FILE, sgr_lock, sgr_unlock);
		else
#endif
			vipwedit (GROUP_FILE, gr_lock, gr_unlock);
	}

	nscd_flush_cache ("passwd");
	nscd_flush_cache ("group");

	return 0;
}
